#include "VHI/Windows/WavAudioRender.h"

#include <windows.h>
#include <mmsystem.h>


static WAVEHDR* AllocateBlocks(uint32 size, int32 count) {
    uint32 totalBufferSize = (size + sizeof(WAVEHDR)) * count;

    uint8* buffer{ nullptr };
    if ((buffer = reinterpret_cast<uint8*>(HeapAlloc(
        GetProcessHeap(),
        HEAP_ZERO_MEMORY,
        totalBufferSize
    ))) == NULL) {
        ExitProcess(1);
    }

    WAVEHDR* blocks = reinterpret_cast<WAVEHDR*>(buffer);
    buffer += sizeof(WAVEHDR) * count;
    for (int32 index = 0; index < count; ++index) {
        blocks[index].dwBufferLength = size;
        blocks[index].lpData = (LPSTR)buffer;
        blocks[index].dwFlags = 0;
        buffer += size;
    }
    return blocks;
}

static void FreeBlocks(WAVEHDR* blockArray) {
    HeapFree(GetProcessHeap(), 0, blockArray);
}


class WavAudioRenderImpl {
public:
    using UtilComplateCallback = std::function<void()>;
public:
    explicit WavAudioRenderImpl(UtilComplateCallback callback) noexcept;
    ~WavAudioRenderImpl();

    bool Write(const char* data, uint32 size);
    static void CALLBACK waveOutProc(HWAVEOUT hwo, UINT uMsg, DWORD_PTR dwInstance,
        DWORD_PTR dwParam1, DWORD_PTR dwParam2);
    void FreeBlock();

private:
    static constexpr int BlockSize_{ 1280 };
    static constexpr int BlockCount_{ 10 };

    uint32 freeBlockCounter_{ BlockCount_ };
    uint32 waveCurrentBlock_{ 0 };

    WAVEHDR* waveBlocks_{ nullptr };
    HWAVEOUT hWavout_{ nullptr };
    bool initialized_{ false };
    UtilComplateCallback callback_;
};

WavAudioRenderImpl::WavAudioRenderImpl(UtilComplateCallback callback) noexcept
    : callback_(callback){
    HWAVEOUT hWaveOut = nullptr;

    waveBlocks_ = AllocateBlocks(BlockSize_, BlockCount_);

    constexpr int32 rate = 16000;
    constexpr int32 channel = 1;
    constexpr int32 bitePerSample = 16;

    WAVEFORMATEX waveform;
    waveform.wFormatTag = WAVE_FORMAT_PCM;
    waveform.nChannels = channel;
    waveform.nSamplesPerSec = rate;
    waveform.nAvgBytesPerSec = rate * channel * 2;
    waveform.nBlockAlign = channel * 2;
    waveform.wBitsPerSample = bitePerSample;
    waveform.cbSize = 0;


    MMRESULT hr = waveOutOpen(&hWaveOut, WAVE_MAPPER, &waveform,
        reinterpret_cast<DWORD_PTR>(WavAudioRenderImpl::waveOutProc),
        reinterpret_cast<DWORD_PTR>(this), CALLBACK_FUNCTION);
    if (MMSYSERR_NOERROR != hr) {
        return;
    }

    hWavout_ = std::move(hWaveOut);
    initialized_ = true;
}

WavAudioRenderImpl::~WavAudioRenderImpl() {
    initialized_ = false;
    if (nullptr == hWavout_) {
        return;
    }
    waveOutReset(hWavout_);
    for (int i = 0; i < BlockCount_; i++) {
        if (waveBlocks_[i].dwFlags & WHDR_PREPARED) {
            waveOutUnprepareHeader(hWavout_, &waveBlocks_[i], sizeof(WAVEHDR));
        }
    }

    do {
        MMRESULT ret = waveOutClose(hWavout_);
        if (MMSYSERR_NOERROR == ret) {
            break;
        } else if (WAVERR_STILLPLAYING == ret) {
            ::Sleep(1);
        } else if (MMSYSERR_NOMEM == ret) {
            break;
        } else {
            break;
        }
    } while (true);
    FreeBlocks(waveBlocks_);
}

bool WavAudioRenderImpl::Write(const char* data, uint32 size) {
    if (nullptr == hWavout_) {
        return false;
    }

    WAVEHDR* current;
    int32 remain;
    current = &waveBlocks_[waveCurrentBlock_];
    while (size > 0) {
        if (current->dwFlags & WHDR_PREPARED) {
            waveOutUnprepareHeader(hWavout_, current, sizeof(WAVEHDR));
        }
        if (size < (int)(BlockSize_ - current->dwUser)) {
            memcpy(current->lpData + current->dwUser, data, size);
            current->dwUser += size;
            break;
        }
        remain = BlockSize_ - static_cast<int32>(current->dwUser);
        memcpy(current->lpData + current->dwUser, data, remain);
        size -= remain;
        data += remain;
        current->dwBufferLength = BlockSize_;
        waveOutPrepareHeader(hWavout_, current, sizeof(WAVEHDR));
        waveOutWrite(hWavout_, current, sizeof(WAVEHDR));
        --freeBlockCounter_;

        while (!freeBlockCounter_ && initialized_) {
            Sleep(1);
        }
        if (!initialized_) {
            break;
        }

        ++waveCurrentBlock_;
        waveCurrentBlock_ %= BlockCount_;
        current = &waveBlocks_[waveCurrentBlock_];
        current->dwUser = 0;
    }
    return true;
}

void CALLBACK WavAudioRenderImpl::waveOutProc(HWAVEOUT hwo, UINT uMsg, DWORD_PTR dwInstance,
    DWORD_PTR dwParam1, DWORD_PTR dwParam2) {
    WavAudioRenderImpl* self = reinterpret_cast<WavAudioRenderImpl*>(dwInstance);

    if (uMsg != WOM_DONE)
        return;

    self->FreeBlock();
}

void WavAudioRenderImpl::FreeBlock() {
    ++freeBlockCounter_;
    if (nullptr != callback_) {
        callback_();
    }
}



WavAudioRender::WavAudioRender() noexcept {
    auto utilComplatedCallback = [this]() {
        ++playedIndex_;
    };

    auto run = [this, utilComplatedCallback]() {
        WavAudioRenderImpl impl(utilComplatedCallback);
        std::vector<int8> temp;
        while (!showExit_.load()) {
            {
                std::unique_lock<std::mutex> lock(mutex_);
                cv_.wait(lock, [&]() {return !audioQueue_.empty() || showExit_.load(); });
                if (showExit_.load()) {
                    break;
                }
                temp = std::move(audioQueue_.front());
                audioQueue_.pop();
            }
            impl.Write(temp.data(), temp.size());
        }
    };

    work_.reset(new std::thread(run));
}

WavAudioRender::~WavAudioRender() {
    showExit_.store(true);
    cv_.notify_one();
    if (work_->joinable()) {
        work_->join();
    }
}

bool WavAudioRender::Write(const char* data, uint32 size) {
    WirteLog(DEBUG, "WavAudioRender::Write");

    std::vector<int8> frame(size, 0);
    memcpy(&frame[0], data, size);
    
    std::unique_lock<std::mutex> lock(mutex_);
    audioQueue_.emplace(std::move(frame));
    cv_.notify_one();
    return true;
}

